#[doc(inline)]
#[cfg(feature = "proc-macro")]
pub use async_main::async_main;

use std::{sync::{mpsc::{SyncSender, self, RecvError}, Arc, Mutex}, pin::Pin, future::Future, task::{Context, Poll}};

pub fn block_on<F: Future>(fut: F) -> Result<F::Output, RecvError> {
    struct Task {
        tx: SyncSender<()>
    }

    impl ArcWake for Task {
        fn wake_by_ref(self: &Arc<Self>) {
            self.tx.send(()).unwrap();
        }
    }
    
    let (tx, rx) = mpsc::sync_channel(100);
    tx.send(()).unwrap();
    let waker = waker(Arc::new(Task { tx }));
    let mut context = Context::from_waker(&waker);
    let mut future = Box::pin(fut);

    loop {
        rx.recv()?;
        match future.as_mut().poll(&mut context) {
            ::core::task::Poll::Pending => continue,
            ::core::task::Poll::Ready(v) => break Ok(v)
        }
    }
}

pub struct Executor {}

impl Executor {
    pub fn new() -> Self {
        Executor {}
    }

    pub fn run(&self, fut: impl Future<Output = ()> + Send + Sync + 'static) {
        let (tx, rx) = mpsc::sync_channel(10000);

        let main = Arc::new(Task { tx: tx.clone(), fut: Mutex::new(Box::pin(fut)) });

        tx.send(main).unwrap();
        drop(tx);

        while let Ok(task) = rx.recv() {
            let waker = waker(task.clone());
            let mut context = Context::from_waker(&waker);
            #[allow(unused_must_use)]
            { task.fut.lock().unwrap().as_mut().poll(&mut context); }
        }
    }
}
pub struct Task {
    tx: SyncSender<Arc<Task>>,
    fut: Mutex<Pin<Box<dyn Future<Output = ()> + Send + Sync>>> // dunno how to fix this, just hope everyhing is send/sync for now
}

impl ArcWake for Task {
    fn wake_by_ref(self: &Arc<Self>) {
        self.tx.send(self.clone()).unwrap();
    }
}

pub mod futures;
use crate::futures::*;

pub mod stream;

pub mod io;

pub mod wake;
use wake::*;

struct NoopWaker;

impl ArcWake for NoopWaker {
    fn wake_by_ref(self: &Arc<Self>) {}
}

pub trait FutureExt: Future + 'static {
    fn poll_unpin(&mut self, cx: &mut Context) -> Poll<Self::Output>
    where
        Self: Unpin
    {
        Pin::new(self).poll(cx)
    }

    fn try_unwrap(self) -> Option<Self::Output>
    where
        Self: Sized
    {
        match Box::pin(self).as_mut().poll(&mut Context::from_waker(&waker(Arc::new(NoopWaker)))) {
            Poll::Ready(v) => Some(v),
            Poll::Pending => None
        }
    }

    fn and<U>(self, other: impl Future<Output = U> + 'static) -> AllFuture<Self::Output, U>
    where
        Self: Sized
    {
        AllFuture::new(self, other)
    }

    fn or<U>(self, other: impl Future<Output = U> + 'static) -> AnyFuture<Self::Output, U>
    where
        Self: Sized
    {
        AnyFuture::new(self, other)
    }
}

impl<F: Future + 'static> FutureExt for F {}

struct FromPollFn<T, F: Fn(&mut Context) -> Poll<T>> {
    f: F
}

impl<T, F: Fn(&mut Context) -> Poll<T>> Future for FromPollFn<T, F> {
    type Output = T;

    fn poll(self: Pin<&mut Self>, cx: &mut Context<'_>) -> Poll<Self::Output> {
        unsafe { self.map_unchecked_mut(|this| &mut this.f)(cx) }
    }
}

pub fn poll_fn<T, F: Fn(&mut Context) -> Poll<T>>(f: F) -> impl Future<Output = T> {
    FromPollFn { f }
}



/// A macro for pin projection.
/// 
/// ```
/// # use messing_with_async::pin_project;
/// # use std::marker::PhantomPinned;
/// struct Thing {
///     field: usize,
///     _phantom: PhantomPinned
/// }
/// 
/// let mut pinned_thing = Box::pin(Thing { field: 5, _phantom: PhantomPinned });
/// 
/// pin_project!(pinned_thing.as_mut() => field);
/// 
/// println!("{}", field);
/// ```
#[macro_export]
macro_rules! pin_project {
    ($self:expr => $field:ident) => {
        let $field = unsafe { ::std::pin::Pin::map_unchecked_mut($self, |this| &mut this.$field) };
    };
    ($self:expr => { $($field:ident),+ }) => {
        let ($($field),+) = unsafe {
            let this = ::std::pin::Pin::get_unchecked_mut($self);
            (
                $(std::pin::Pin::new_unchecked(&mut this.$field)),+
            )
        };
    };
    (mut $self:expr => $field:ident) => {
        let mut $field = unsafe { ::std::pin::Pin::map_unchecked_mut($self, |this| &mut this.$field) };
    };
    (mut $self:expr => { $($field:ident),+ }) => {
        let ($(mut $field),+) = unsafe {
            let this = ::std::pin::Pin::get_unchecked_mut($self);
            (
                $(std::pin::Pin::new_unchecked(&mut this.$field)),+
            )
        };
    };
}