use std::{pin::Pin, task::{Context, Poll}, future::Future};

use crate::pin_project;

/// An asynchronous version of [`Iterator`].
/// 
/// There exists an [`std::stream::Stream`], but it is unstable.
/// 
/// ```
/// # use messing_with_async::{async_main, stream::{StreamExt, Stream, from_iter}};
/// # async_main!(async_main);
/// # async fn async_main() {
/// let mut stream = from_iter(vec![1, 2, 3]);
/// assert_eq!(stream.next().await, Some(1));
/// assert_eq!(stream.next().await, Some(2));
/// assert_eq!(stream.next().await, Some(3));
/// assert_eq!(stream.next().await, None);
/// # }
/// # main();
/// ```
#[must_use]
pub trait Stream {
    type Item;

    fn poll_next(self: Pin<&mut Self>, cx: &mut Context<'_>) -> Poll<Option<Self::Item>>;

    fn size_hint(&self) -> (usize, Option<usize>) {
        (0, None)
    }
}

pub trait StreamExt: Stream {
    /// Creates a future which produces the next element of the stream.
    /// 
    /// ```
    /// # use messing_with_async::{async_main, stream::{StreamExt, Stream, from_iter}};
    /// # async_main!(async_main);
    /// # async fn async_main() {
    /// let mut stream = from_iter(vec![1, 2, 3]);
    /// assert_eq!(stream.next().await, Some(1));
    /// assert_eq!(stream.next().await, Some(2));
    /// assert_eq!(stream.next().await, Some(3));
    /// assert_eq!(stream.next().await, None);
    /// # }
    /// # main();
    /// ```
    fn next(&mut self) -> StreamNextFuture<Self> {
        StreamNextFuture {
            stream: self
        }
    }

    /// Transforms the [`Stream`] into a future which evalates to an [`Iterator`] of `Self::Item`.
    /// 
    /// ```
    /// # use messing_with_async::{async_main, stream::{StreamExt, Stream, from_iter}};
    /// # async_main!(async_main);
    /// # async fn async_main() {
    /// let stream = from_iter(vec![1, 2, 3]);
    /// let iter = stream.into_iter_fut().await;
    /// assert_eq!(iter.collect::<Vec<_>>(), vec![1, 2, 3])
    /// # }
    /// # main();
    /// ```
    fn into_iter_fut(self) -> StreamIterFuture<Self>
    where
        Self: Sized
    {
        StreamIterFuture {
            stream: self,
            vec: Vec::new()
        }
    }

    fn advance_by(&mut self, n: usize) -> StreamAdvanceByFuture<Self>
    where
        Self: Sized
    {
        StreamAdvanceByFuture::new(self, n)
    }

    /// Creates a new stream which will produce up to `count` items from the original stream.
    /// 
    /// ```
    /// # use messing_with_async::{async_main, stream::{StreamExt, Stream, from_iter}};
    /// # async_main!(async_main);
    /// # async fn async_main() {
    /// let stream = from_iter(vec![1, 2, 3, 4, 5]);
    /// let mut new_stream = stream.take(3);
    /// assert_eq!(new_stream.next().await, Some(1));
    /// assert_eq!(new_stream.next().await, Some(2));
    /// assert_eq!(new_stream.next().await, Some(3));
    /// assert_eq!(new_stream.next().await, None);
    /// # }
    /// # main();
    /// ```
    fn take(self, count: usize) -> Take<Self>
    where
        Self: Sized
    {
        Take::new(self, count)
    }
}

impl<T: Stream + ?Sized> StreamExt for T {}



/// A future which produces the next item of a [`Stream`].
pub struct StreamNextFuture<'a, S: Stream + ?Sized> {
    stream: &'a mut S
}

impl<'a, S: Stream + ?Sized> StreamNextFuture<'a, S> {
    pub fn new(stream: &'a mut S) -> Self {
        Self { stream }
    }
}

impl<S: Stream + ?Sized> Future for StreamNextFuture<'_, S> {
    type Output = Option<S::Item>;

    fn poll(self: Pin<&mut Self>, cx: &mut Context<'_>) -> Poll<Self::Output> {
        let stream = unsafe { self.map_unchecked_mut(|this| this.stream) };
        stream.poll_next(cx)
    }
}



/// A future which transforms a [`Stream`] into an [`Iterator`].
pub struct StreamIterFuture<S: Stream> {
    stream: S,
    vec: Vec<S::Item>
}

impl<S: Stream> StreamIterFuture<S> {
    pub fn new(stream: S) -> Self {
        Self { stream, vec: Vec::new() }
    }
}

impl<S: Stream> Future for StreamIterFuture<S> {
    type Output = std::vec::IntoIter<S::Item>;

    fn poll(self: Pin<&mut Self>, cx: &mut Context<'_>) -> Poll<Self::Output> {
        pin_project!(mut self => {stream, vec});

        loop {
            match stream.as_mut().poll_next(cx) {
                Poll::Ready(Some(v)) => unsafe { vec.as_mut().get_unchecked_mut() }.push(v),
                Poll::Ready(None) => {
                    let new = unsafe { vec.get_unchecked_mut() }.drain(..).collect::<Vec<_>>();
                    break Poll::Ready(new.into_iter())
                },
                Poll::Pending => break Poll::Pending,
            }
        }
    }
}



/// A [`Stream`] which can be created from an [`Iterator`].
pub struct FromIter<I: Iterator> {
    iter: I
}

/// Transforms an [`Iterator`] into a [`Stream`].
/// 
/// ```
/// # use messing_with_async::{async_main, stream::{StreamExt, Stream, from_iter}};
/// # async_main!(async_main);
/// # async fn async_main() {
/// let mut stream = from_iter(vec![1, 2, 3]);
/// assert_eq!(stream.next().await, Some(1));
/// assert_eq!(stream.next().await, Some(2));
/// assert_eq!(stream.next().await, Some(3));
/// assert_eq!(stream.next().await, None);
/// # }
/// # main();
/// ```
pub fn from_iter<I: IntoIterator>(iter: I) -> FromIter<I::IntoIter> {
    FromIter { iter: iter.into_iter() }
}

impl<I: Iterator> Stream for FromIter<I> {
    type Item = I::Item;

    fn poll_next(self: Pin<&mut Self>, _: &mut Context<'_>) -> Poll<Option<Self::Item>> {
        Poll::Ready(unsafe { self.get_unchecked_mut() }.iter.next())
    }
}



/// A future which produces the next item of a [`Stream`].
pub struct StreamAdvanceByFuture<'a, S: Stream + ?Sized> {
    stream: &'a mut S,
    n: usize
}

impl<'a, S: Stream> StreamAdvanceByFuture<'a, S> {
    pub fn new(stream: &'a mut S, n: usize) -> Self {
        Self { stream, n }
    }
}

impl<S: Stream> Future for StreamAdvanceByFuture<'_, S> {
    type Output = Result<(), usize>;

    fn poll(self: Pin<&mut Self>, cx: &mut Context<'_>) -> Poll<Self::Output> {
        let n = self.n;
        let mut stream = unsafe { self.map_unchecked_mut(|this| this.stream) };
        for i in 0..n {
            match stream.as_mut().poll_next(cx) {
                Poll::Ready(Some(_)) => {},
                Poll::Ready(None) => return Poll::Ready(Err(i)),
                Poll::Pending => return Poll::Pending
            }
        }
        Poll::Ready(Ok(()))
    }
}




/// A [`Stream`] which produces up to a given amount of items from another [`Stream`].
/// 
/// See also:
/// [`std::iter::Take`]
/// [`StreamExt::take`]
pub struct Take<S: Stream> {
    stream: S,
    count: usize
}

impl<S: Stream> Take<S> {
    pub fn new(stream: S, count: usize) -> Self {
        Self { stream, count }
    }
}

impl<S: Stream> Stream for Take<S> {
    type Item = S::Item;

    fn poll_next(self: Pin<&mut Self>, cx: &mut Context<'_>) -> Poll<Option<Self::Item>> {
        let (stream, count) = unsafe {
            let this = self.get_unchecked_mut();
            (Pin::new_unchecked(&mut this.stream), &mut this.count)
        };

        if *count == 0 {
            return Poll::Ready(None);
        }

        let res = stream.poll_next(cx);

        if let Poll::Ready(Some(_)) = res {
            *count -= 1;
        }

        res
    }
}