use std::{task::{Waker, RawWakerVTable, RawWaker}, sync::Arc, mem::ManuallyDrop};

pub trait ArcWake {
    fn wake_by_ref(self: &Arc<Self>);

    fn wake(self: Arc<Self>) {
        ArcWake::wake_by_ref(&self)
    }
}

pub fn waker<W: ArcWake>(arc_wake: Arc<W>) -> Waker {
    let vtable = vtable::<W>();
    let raw_waker = RawWaker::new(Arc::into_raw(arc_wake) as *const (), vtable);
    unsafe { Waker::from_raw(raw_waker) }
}

fn vtable<W: ArcWake>() -> &'static RawWakerVTable {
    &RawWakerVTable::new(
        |ptr| unsafe {
            // increment the strong count and make a clone
            Arc::increment_strong_count(ptr);
            RawWaker::new(ptr, vtable::<W>())
        },
        |ptr| unsafe {
            // here the arc should be dropped because we are supposed to consume the ptr
            ArcWake::wake(Arc::from_raw(ptr as *const W))
        },
        |ptr| unsafe {
            // must make sure to not drop the arc, so as to not decrement strong count
            ArcWake::wake_by_ref(&*ManuallyDrop::new(Arc::from_raw(ptr as *const W)))
        },
        |ptr| unsafe {
            // will get dropped, decrementing strong count
            Arc::from_raw(ptr as *const W);
        }
    )
}
